/* -*-ObjC-*-
*******************************************************************************
*
* File:         MailboxSummaryList.h
* RCS:          MailboxSummaryList.h,v 1.7 1998/06/08 20:50:34 tom Exp
* Description:  
* Author:       Tom Hageman <tom@basil.icce.rug.nl>
* Created:      April 1998
* Modified:     
* Language:     Objective-C
* Package:      EnhanceMail
* Status:       Experimental
*
* Copyright (C) 1998 Tom Hageman, but otherwise this file is perfect freeware.
*
*******************************************************************************
*/
#import <objc/List.h>

#import "MailboxSummary.h"


@interface EnhanceMailboxSummaryList : List
{
   char *fullname;
   enum EMSLoadStatus loadStatus;
   BOOL _dontScan;
   BOOL _doScan;
   BOOL _suppressed;
   BOOL _autoUpdateCache;
   // Incremental loadig support.
   id iterateLoadContext;
   id iterateLoadEnumerator;
}

+ summaryListForName:(const char *)name;
- initWithName:(const char *)name;
- (const char *)name;
- (void)setName:(const char *)name;
- (const char *)shortName;

// Access methods.
- (BOOL)dontScan;
- (BOOL)doScan;
- (void)setDontScan:(BOOL)aValue;
- (void)setDoScan:(BOOL)aValue;
   // These are mutually exclusive.  If both are unset, inherit from parent.

- (BOOL)autoUpdateCache;
- (void)setAutoUpdateCache:(BOOL)aValue;

// Loading.
- (enum EMSLoadStatus)load;
- (enum EMSLoadStatus)loadUnless:(BOOL)suppressed;
- (int)scandir;

- (enum EMSLoadStatus)loadStatus;
- (BOOL)isLoaded;
- (BOOL)isLoadSuppressed;

// Support.
- (const char *)getPath:(char *)buf forFile:(const char *)file;

// General results.
- (long)count:(int)tag;
- (unsigned long)size:(int)tag;
- (long)attachCount:(int)tag;
- (unsigned long)attachSize:(int)tag;

// Specialized results.
- (long)numUnreadMessages;
- (long)numNewMessages;

- (int)numMailboxesWithNewOrUnreadMessages;

- (BOOL)hasPendingAppnmail;

- (BOOL)hasNewMail;
- (int)numMailboxesWithNewMail;

@end // EnhanceMailboxSummaryList


@interface EnhanceMailboxSummaryList (Caching)

+ cachedSummaryListForName:(const char *)name;
+ (BOOL)storeSummaryListToCache:(EnhanceMailboxSummaryList *)list;
- (BOOL)storeToCache;

@end // EnhanceMailboxSummaryList (Caching)


@interface EnhanceMailboxSummaryList (IterateLoad)

+ (int)iterateLoadInterval;
+ (void)setIterateLoadInterval:(int)value;

- (int)iterateLoad:sender;
- (BOOL)iterating;
- (void)iterateAbort;

@end // EnhanceMailboxSummaryList (IterateLoad)


@interface Object (EnhanceIterateLoadCallbackMethods)
- (void)didFinishIterateLoad:(EnhanceMailboxSummaryList *)sender;
- (void)newMailForSummary:(EnhanceMailboxSummary *)sender;
@end
